/**
 * ********************************************************************
 *
 * Application = 
 *
 * Source  = 
 *
 * Objet   = Hash table library 
 *
 * --------------------------------------------------------------------
 *
 * Remarques =
 *
 * --------------------------------------------------------------------
 *
 * Auteur  = Gregoire Lejeune ; Version 1.0 ; 03-11-2000
 *
 * ********************************************************************
 */
 
#define _USES_HASH_TABLE_
#define _USES_LHASH_

#include "LHash.h"

/**
 * ********************************************************************
 *
 * BeginSection = Declaration d'elements globaux pour le module
 *
 */
 
#define __MODULE__  "LHash"
 
#ifndef True
 #define True 1
#endif
 
#ifndef False
 #define False 0
#endif
 
#define __HASHSTART    5381
#define __BIG_CAPACITY 240101
#define __EMPTYISH     0.05
#define __GROW_DELTA   2
#define __SHRINK_DELTA __GROW_DELTA
#define ___FULLHASH    0.60

static const uint32_t liPrimeList[] = {
  11,      19,      37,      73,      109,      163,      251,
  367,     557,     823,     1237,    1861,     2777,     4177,
  6247,    9371,    14057,   21089,   31627,    47431,    71143,
  106721,  160073,  240101,  360163,  540217,   810343,   1215497,
  1823231, 2734867, 4102283, 6153409, 9230113,  13845163,
};

static size_t szPrimeList = sizeof(liPrimeList) / sizeof(liPrimeList[0]);

static uint32_t LHHashAdd( uint32_t, uint8_t );
static uint32_t LHHashFunc( const char *, size_t );
static void LHDestroyNode( THashNode * );
static uint32_t Prime( uint32_t );
static void LHShrink( THashTable * );
static void LHGrow( THashTable * );
static void LHResize( THashTable *, int );

/**
 *
 * EndSection = Declaration D'Elements Globaux Pour Le Module
 *
 * ********************************************************************
 */
 
/**
 * ********************************************************************
 *
 * BeginSection = Services executifs du module
 *
 */
 
/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
THashTable * _LHInitHash( size_t iCapacity )
{
 #define __ENTRY__ __MODULE__"._LHInitHash()"
 
 int iCpt;
 THashTable *ptHashTable = NULL;
 THashNode **pptNodes;

 if( !iCapacity )
  { iCapacity = __DEFAULT_CAPACITY; }

 ptHashTable = (THashTable *) malloc(sizeof(THashTable));
 if( !ptHashTable )
   { goto Finalize; }
 
 ptHashTable->iCapacity  = iCapacity-1;
 ptHashTable->fFullSize  = ___FULLHASH;
 ptHashTable->iNbEntries = 0;
 ptHashTable->ptCurrentSearchNode = NULL;
 ptHashTable->iCurrentSearchNode = -1;

 if( !( pptNodes = malloc( sizeof(THashNode *) * iCapacity ) ) )
 {
	free( ptHashTable );
	ptHashTable = NULL;
    goto Finalize;
 }
 for( iCpt = 0; iCpt < iCapacity; ++iCpt )
  { pptNodes[iCpt] = NULL; }
 
 ptHashTable->pptNodes = pptNodes;    
 
 Finalize:
 return ptHashTable;
 
 #undef __ENTRY__
}
/**
 * ********************************************************************
 *
 * Service = _LHFreeHash
 *
 * Tache   = Libere la table de hashage
 *
 * --------------------------------------------------------------------
 *
 * Entrees =    pptHashTable : pointeur sur pointeur sur la table de hashage
 *                        callback : pointeur sur la fonction de liberation de la structure
 *                        rattachee a la table.
 *
 * Retour  = code d'erreur 0 si OK, -1 sinon
 *
 * ********************************************************************
 */
void _LHFreeHash(THashTable * pptHashTable, void (*callback) (void *))
{
 #define __ENTRY__ __MODULE__"._LHFreeHash()"
	
 int iCpt;
 THashNode *ptNextNode;

 for( iCpt=0; iCpt<(pptHashTable)->iCapacity; iCpt++ )
 {
   if( (pptHashTable)->pptNodes[iCpt] != NULL )
   {
      free( (pptHashTable)->pptNodes[iCpt]->pxKey );
	  (pptHashTable)->pptNodes[iCpt]->pxKey = NULL;
      if( pptHashTable->pptNodes[iCpt]->pvData != NULL && callback != NULL) {
		(*callback)( pptHashTable->pptNodes[iCpt]->pvData );
		pptHashTable->pptNodes[iCpt]->pvData = NULL;
	  }

      while( pptHashTable->pptNodes[iCpt]->ptNextNode != NULL ) 
      {
        ptNextNode = pptHashTable->pptNodes[iCpt]->ptNextNode;
        pptHashTable->pptNodes[iCpt]->ptNextNode = pptHashTable->pptNodes[iCpt]->ptNextNode->ptNextNode;
        free( ptNextNode->pxKey );
		ptNextNode->pxKey = NULL;
        if( ptNextNode->pvData != NULL && callback != NULL ) {
		  (*callback)( ptNextNode->pvData );
		  ptNextNode->pvData = NULL;
		}
      }
      free( pptHashTable->pptNodes[iCpt] );
	  pptHashTable->pptNodes[iCpt] = NULL;
   }
 }
 free( pptHashTable->pptNodes );
 pptHashTable->pptNodes = NULL;
 free( pptHashTable );
 pptHashTable = NULL;
 
 return;
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
void *_LHRemoveValue( THashTable *ptHashTable, const char *pxKey, size_t iKeyLength )
{
 #define __ENTRY__ __MODULE__"._LHRemoveValue()"

 void *RCod = NULL;

 /** Var = Tableau des noeuds de la table */
 THashNode **pptNodes = ptHashTable->pptNodes;
 /** Var = Noeud courant */
 THashNode *ptCurrentNode;
 /** Var = Noeud a supprimer */
 THashNode *ptRemovedNode;
 
 /** Act = Calcul de la cle de hashage */
 uint32_t iHashValue = LHHashFunc( pxKey, iKeyLength ) % ptHashTable->iCapacity;
 
 /** Act = On verifie si la cle marche */
 if( !( ptCurrentNode = pptNodes[iHashValue] ) )
  { goto Finalize; }

 /** Act = Suppression */
 if( !strcmp( ptCurrentNode->pxKey, pxKey ) ) {
   RCod = ptCurrentNode->pvData;
   pptNodes[iHashValue] = ptCurrentNode->ptNextNode;
   LHDestroyNode( ptCurrentNode );
   free( ptCurrentNode );
   ptCurrentNode = NULL;
   --(ptHashTable->iNbEntries);
 } else {
   for( ; ( ptRemovedNode = ptCurrentNode->ptNextNode ); ptCurrentNode = ptCurrentNode->ptNextNode ) {
     if( !strcmp( pxKey, ptRemovedNode->pxKey ) ) {
       RCod = ptRemovedNode->pvData;
       ptCurrentNode->ptNextNode = ptRemovedNode->ptNextNode;
       LHDestroyNode( ptRemovedNode );
       free( ptRemovedNode );
	   ptRemovedNode = NULL;
       --(ptHashTable->iNbEntries);
       break;
     }
   }
 }

 /** Act = Reorganisation des donnes - si besoin */
 if( ( ptHashTable->iCapacity > __BIG_CAPACITY ) && 
     ( ( (float)(ptHashTable->iNbEntries / ptHashTable->iCapacity) ) < __EMPTYISH ) )
  { LHShrink( ptHashTable ); }

 Finalize:
 return RCod;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
void LHInitSearch( THashTable *ptHashTable )
{
 #define __ENTRY__ __MODULE__".LHInitSearch()"
	
 ptHashTable->iCurrentSearchNode = 0;
 
 return;

 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
THashNode *LHSearchNext( THashTable *ptHashTable )
{
 #define __ENTRY__ __MODULE__".LHSearchNext()"
	
  /** Var: Noeud resultat de la recherche */
 THashNode *ptCurrentNode = NULL;
 
  /** Act: On initialise la recherche si ce n'est pas deja fait */
 if( ptHashTable->iCurrentSearchNode < 0 ) 
   { LHInitSearch( ptHashTable ); }
 
  /** Rem: Il n'y a plus de noeuds a parcourir, on renvoie NULL */
 if( ptHashTable->iCurrentSearchNode >= ptHashTable->iCapacity &&
     ptCurrentNode != NULL &&
     ptCurrentNode->ptNextNode == NULL )
   { goto Finalize; }
   
  /** Act: Recherche du prochain noeud de meme niveau */
 if( ptCurrentNode != NULL ) {
   if( ptCurrentNode->ptNextNode != NULL ) {
     ptCurrentNode = ptCurrentNode->ptNextNode;
     goto Finalize;
   }
 }

  /** Act: Recherche du prochain noeud */
 while ( ptHashTable->pptNodes[ptHashTable->iCurrentSearchNode] == NULL &&
		 ptHashTable->iCurrentSearchNode < ptHashTable->iCapacity )
   { ptHashTable->iCurrentSearchNode++; }

  /** Act: Recuperation de la donnes */
 if( ptHashTable->iCurrentSearchNode < ptHashTable->iCapacity ) {
   ptCurrentNode = ptHashTable->pptNodes[ptHashTable->iCurrentSearchNode];
   ptHashTable->iCurrentSearchNode++;
 }

 Finalize:
 return( ptCurrentNode );

 #undef __ENTRY__
}
/**
 *
 * EndSection   = Services executifs du module
 *
 * ********************************************************************
 */
 
/**
 * ********************************************************************
 *
 * BeginSection = Services de type set du module
 *
 */
 
/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
long _LHSetValue( THashTable *ptHashTable, const char *pxKey, size_t iKeyLength, void *pvData )
{
 #define __ENTRY__ __MODULE__"._LHSetValue()"

 int RCod = True;
 
 void *replaced = NULL;
 THashNode **pptNodes = ptHashTable->pptNodes;
 THashNode *ptCurrentNode;
 uint32_t iHashValue = LHHashFunc(pxKey, iKeyLength) % ptHashTable->iCapacity;

 for( ptCurrentNode = pptNodes[iHashValue]; ptCurrentNode; ptCurrentNode = ptCurrentNode->ptNextNode ) 
 {
   if( !strcmp( pxKey, ptCurrentNode->pxKey ) ) {
     replaced = ptCurrentNode->pvData;
     ptCurrentNode->pvData = pvData;
     break;
   }
 }

 if( !replaced ) 
 {
   if( !( ptCurrentNode = malloc( sizeof(THashNode) ) ) )
   {
     RCod = False;
     goto Finalize;
   }
   if( !( ptCurrentNode->pxKey = strdup( pxKey ) ) )
   {
     RCod = False;
     goto Finalize;
   }
   ptCurrentNode->ptNextNode = pptNodes[iHashValue];
   pptNodes[iHashValue] = ptCurrentNode;
   ptCurrentNode->pvData = pvData;
   ptCurrentNode->iKeyLength = iKeyLength;
   ++(ptHashTable->iNbEntries);

   if( ( (float)(ptHashTable->iNbEntries / ptHashTable->iCapacity) ) > ptHashTable->fFullSize )
    { LHGrow( ptHashTable ); }
 }

 Finalize:
 return RCod;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
void _LHSetFull( THashTable *ptHashTable, float fFullSize )
{
 #define __ENTRY__ __MODULE__"._LHSetValueFullHash()"

 if( fFullSize > 1 ) {
   ptHashTable->fFullSize = fFullSize / 100;
 } else {
   ptHashTable->fFullSize = fFullSize;
 }

 return;
 
 #undef __ENTRY__
}

/**
 *
 * EndSection   = Services de type set du module
 *
 * ********************************************************************
 */
 
/**
 * ********************************************************************
 *
 * BeginSection = Services de type get du module
 *
 */
 
/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
void *_LHGetValue( THashTable *ptHashTable, const char *pxKey, size_t iKeyLength )
{
 #define __ENTRY__ __MODULE__"._LHGetValue()"

 void *RCod = NULL;
  
 /** Var: Noeud courant */
 THashNode *ptCurrentNode;
 /** Var: Table des noeuds */
 THashNode **pptHashNodeTable = ptHashTable->pptNodes;

 for( ptCurrentNode = pptHashNodeTable[LHHashFunc( pxKey, iKeyLength ) % ptHashTable->iCapacity]; 
      ptCurrentNode; ptCurrentNode = ptCurrentNode->ptNextNode )
 {
   if( !strcmp( pxKey, ptCurrentNode->pxKey ) )
   {
     RCod = ptCurrentNode->pvData;
     goto Finalize;
   }
 }
 
 Finalize:
 return RCod;
 
 #undef __ENTRY__
}

/**
 *
 * EndSection   = Services de type get du module
 *
 * ********************************************************************
 */

/**
 * ********************************************************************
 *
 * BeginSection = Services prives du module 
 *
 */
 
/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static uint32_t LHHashAdd(uint32_t iHashValue, uint8_t iDataChar)
{
 #define __ENTRY__ __MODULE__".LHHashAdd()"
    
 iHashValue += (iHashValue << 5);
 return( iHashValue ^ iDataChar );
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static uint32_t LHHashFunc( const char *pxKey, size_t iKeyLength )
{
 #define __ENTRY__ __MODULE__".LHHashFunc()"
    
 uint32_t iHashValue;

 iHashValue = __HASHSTART;
 while( iKeyLength ) {
   iHashValue = LHHashAdd( iHashValue, *pxKey++ );
   --iKeyLength;
 }
 
 return iHashValue;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static void LHDestroyNode( THashNode *ptNode )
{
 #define __ENTRY__ __MODULE__".LHDestroyNode()"
 
 free( ptNode->pxKey );
 ptNode->pxKey = NULL;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static uint32_t Prime(uint32_t iPrimeNumber)
{
 extern const uint32_t liPrimeList[];
 extern size_t szPrimeList;
 
 int32_t iCpt;

 for( iCpt = 0; iCpt < szPrimeList; iCpt++ )
 {
   if( liPrimeList[iCpt] > iPrimeNumber )
    { return liPrimeList[iCpt]; }
 }
 
 return liPrimeList[szPrimeList - 1];
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static void LHShrink( THashTable *ptHashTable )
{
 #define __ENTRY__ __MODULE__".LHShrink()"
 
 LHResize( ptHashTable, Prime(ptHashTable->iCapacity / __SHRINK_DELTA) );
 return;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static void LHGrow( THashTable *ptHashTable )
{
 #define __ENTRY__ __MODULE__".LHGrow()"

 LHResize( ptHashTable, Prime(ptHashTable->iCapacity * __GROW_DELTA) );
 return;
 
 #undef __ENTRY__
}

/**
 * ********************************************************************
 *
 * Service = 
 *
 * Tache   = 
 *
 * --------------------------------------------------------------------
 *
 * Entrees = 
 *
 * Retour  = 
 *
 * ********************************************************************
 */
static void LHResize( THashTable *ptHashTable, int iNewCapacity )
{
 #define __ENTRY__ __MODULE__".LHResize()"
 
 size_t iOldCapacity  = ptHashTable->iCapacity;
 THashNode **ptOldNodes = ptHashTable->pptNodes;
 THashNode **ptNewNodes;
 THashNode *ptCurrentNode;
 THashNode *ptOldToNewNode;
 int iCpt;
 uint32_t iHashValue;

 if( !( ptNewNodes = malloc( sizeof(THashNode *) * iNewCapacity ) ) ) {
   goto Finalize;
 }
 
 for( iCpt = 0; iCpt < iNewCapacity; ++iCpt )
  { ptNewNodes[iCpt]  = NULL; }
 
 ptHashTable->pptNodes = ptNewNodes;
 ptHashTable->iCapacity  = iNewCapacity;

 for( iCpt = 0; iCpt < iOldCapacity; ++iCpt ) 
 {
   for( ptCurrentNode = ptOldNodes[iCpt]; ptCurrentNode; ptCurrentNode = ptOldToNewNode ) 
   {
     ptOldToNewNode = ptCurrentNode->ptNextNode;
     iHashValue = LHHashFunc( ptCurrentNode->pxKey, ptCurrentNode->iKeyLength ) % iNewCapacity;
     ptCurrentNode->ptNextNode = ptNewNodes[iHashValue];
     ptNewNodes[iHashValue] = ptCurrentNode;
   }
 }
 free( ptOldNodes );
 ptOldNodes = NULL;

 Finalize:
 return;

 #undef __ENTRY__
}
/**
 *
 * EndSection = Services prives du module 
 *
 * ********************************************************************
 */
