#include <string.h>
#include "graphviz.h"

void ruby_graphviz_mark( TRbGraphViz * );
void ruby_graphviz_free( TRbGraphViz * );

VALUE ruby_graphviz_new2( VALUE class, VALUE xName, VALUE hvConfig ) {
	TRbGraphViz *ptRbGraphViz;
	
	if( rb_obj_is_kind_of( xName, rb_cString ) == Qfalse )
		rb_raise( rb_eTypeError, "Invalid argument: argument #1 must be of type String" );
	if( rb_obj_is_kind_of( hvConfig, rb_cHash ) == Qfalse )
		rb_raise( rb_eTypeError, "Invalid argument: argument #2 must be of type Hash" );

	ptRbGraphViz = (TRbGraphViz *)malloc(sizeof(TRbGraphViz));
	if( ptRbGraphViz == NULL )
		rb_raise(rb_eNoMemError, "No memory left for GraphViz struct");

	/** Act: layout config */
	ptRbGraphViz->eLayout = rb_hash_aref( hvConfig, rb_str_new2( "layout" ) );
	if( ptRbGraphViz->eLayout == Qnil ) { ptRbGraphViz->eLayout = rb_str_new2( "dot" ); }
	
	/** Act: type config */
	ptRbGraphViz->eType   = rb_hash_aref( hvConfig, rb_str_new2( "type" ) );
	if( ptRbGraphViz->eType == Qnil ) { ptRbGraphViz->eType = INT2NUM(AGRAPH); }

	/** Act: output config */
	ptRbGraphViz->xOutput = rb_hash_aref( hvConfig, rb_str_new2( "output" ) );
	if( ptRbGraphViz->xOutput == Qnil ) { ptRbGraphViz->xOutput = INT2NUM(RUBY_GRAPHVIZ_DOT); }
	
	/** Act: Set initialize config */
	ptRbGraphViz->args[0] = strdup( "rgv" );
    ptRbGraphViz->args[1] = (char*)malloc( sizeof(char) * ( strlen( STR2CSTR( ptRbGraphViz->xOutput ) ) + 3 ) );
    sprintf( ptRbGraphViz->args[1], "-T%s", STR2CSTR( ptRbGraphViz->xOutput ) );
	dotneato_initialize( 2, ptRbGraphViz->args );

	/** Act: Open graph */
    ptRbGraphViz->g = agopen( STR2CSTR( xName ), NUM2INT( ptRbGraphViz->eType ) );
	
	ptRbGraphViz->iNbNodes = 0;
	ptRbGraphViz->ptHashNodes = LHInitHash( 5 );
	ptRbGraphViz->iNbEdges = 0;
	ptRbGraphViz->ptHashEdges = LHInitHash( 5 );
	
	return( Data_Wrap_Struct( class, ruby_graphviz_mark, ruby_graphviz_free, ptRbGraphViz ) );
}

void ruby_graphviz_mark( TRbGraphViz *ptRbGraphViz ) {
}

void ruby_graphviz_free( TRbGraphViz *ptRbGraphViz ) {
	switch( NUM2INT(ptRbGraphViz->eLayout) ) {
	  case RUBY_GRAPHVIZ_NEATO:
	    neato_cleanup( ptRbGraphViz->g );
	    break;
	  case RUBY_GRAPHVIZ_TWOPI:
	    twopi_cleanup( ptRbGraphViz->g );
	    break;
	  case RUBY_GRAPHVIZ_DOT:
	  default:
	    dot_cleanup( ptRbGraphViz->g );
    }
	
	/** Act: Close graph */
	agclose( ptRbGraphViz->g );
	
	/** Act: Free GraphViz structure */
	LHFreeHash( ptRbGraphViz->ptHashNodes, NULL );
	LHFreeHash( ptRbGraphViz->ptHashEdges, NULL );
	free( ptRbGraphViz->args[0] );
	free( ptRbGraphViz->args[1] );
	if( ptRbGraphViz != NULL )
		free( ptRbGraphViz );
}

VALUE ruby_graphviz_add_node( VALUE self, VALUE xName, VALUE xNode ) {
	TRbGraphViz *ptRbGraphViz;
	Agnode_t *ptNode;

	if( rb_obj_is_kind_of( xNode, rb_cString ) == Qfalse )
		rb_raise( rb_eTypeError, "Invalid argument: argument #1 must be of type String" );
	
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

    ptNode = agnode( ptRbGraphViz->g, STR2CSTR( xNode ) );
	ptRbGraphViz->iNbNodes += 1;
	LHSetValue( ptRbGraphViz->ptHashNodes, STR2CSTR( xName ), (void *)ptNode );

	return( Qtrue );
}

VALUE ruby_graphviz_add_edge( VALUE self, VALUE xName, VALUE xNode1, VALUE xNode2 ) {
	TRbGraphViz *ptRbGraphViz;
	Agnode_t *ptNode1, *ptNode2;
	Agedge_t *ptEdge;
	
	if( rb_obj_is_kind_of( xNode1, rb_cString ) == Qfalse )
		rb_raise( rb_eTypeError, "Invalid argument: argument #1 must be of type String" );
	if( rb_obj_is_kind_of( xNode2, rb_cString ) == Qfalse )
		rb_raise( rb_eTypeError, "Invalid argument: argument #2 must be of type String" );

	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

    ptNode1 = (Agnode_t *)LHGetValue( ptRbGraphViz->ptHashNodes, STR2CSTR( xNode1 ) );
	ptNode2 = (Agnode_t *)LHGetValue( ptRbGraphViz->ptHashNodes, STR2CSTR( xNode2 ) );
    // ptNode1 = agnode( ptRbGraphViz->g, STR2CSTR( xNode1 ) );
	// ptNode2 = agnode( ptRbGraphViz->g, STR2CSTR( xNode2 ) );
	ptEdge  = agedge( ptRbGraphViz->g, ptNode1, ptNode2 );

	ptRbGraphViz->iNbEdges += 1;
	LHSetValue( ptRbGraphViz->ptHashEdges, STR2CSTR( xName ), (void *)ptEdge );

	return( Qtrue );
}

/** ------------------------------------------------------------------------ */

VALUE ruby_graphviz_add_node_attr( VALUE self, VALUE xAttrName, VALUE xAttrValue ) {
	TRbGraphViz *ptRbGraphViz;
	Agsym_t *ptAttr;
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

	ptAttr = agnodeattr( ptRbGraphViz->g, STR2CSTR( xAttrName ), STR2CSTR( xAttrValue ) );
	
	return( Qtrue );
}

VALUE ruby_graphviz_set_node_attr( VALUE self, VALUE xNodeName, VALUE xAttrName, VALUE xAttrValue ) {
	TRbGraphViz *ptRbGraphViz;
	Agnode_t *ptNode;
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

    ptNode = (Agnode_t *)LHGetValue( ptRbGraphViz->ptHashNodes, STR2CSTR( xNodeName ) );
	agset( ptNode, STR2CSTR( xAttrName ), STR2CSTR( xAttrValue ) );
	
	return( Qtrue );
}

VALUE ruby_graphviz_add_edge_attr( VALUE self, VALUE xAttrName, VALUE xAttrValue ) {
	TRbGraphViz *ptRbGraphViz;
	Agsym_t *ptAttr;
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

	ptAttr = agedgeattr( ptRbGraphViz->g, STR2CSTR( xAttrName ), STR2CSTR( xAttrValue ) );
	
	return( Qtrue );
}

VALUE ruby_graphviz_set_edge_attr( VALUE self, VALUE xEdgeName, VALUE xAttrName, VALUE xAttrValue ) {
	TRbGraphViz *ptRbGraphViz;
	Agedge_t *ptEdge;
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

    ptEdge = (Agedge_t *)LHGetValue( ptRbGraphViz->ptHashEdges, STR2CSTR( xEdgeName ) );
	agset( ptEdge, STR2CSTR( xAttrName ), STR2CSTR( xAttrValue ) );
	
	return( Qtrue );
}

/** ------------------------------------------------------------------------ */

VALUE ruby_graphviz_output( VALUE self ) {
	TRbGraphViz *ptRbGraphViz;
	Data_Get_Struct( self, TRbGraphViz, ptRbGraphViz );

	if( ptRbGraphViz->iNbNodes > 0 || ptRbGraphViz->iNbEdges > 0 ) {
	  switch( NUM2INT(ptRbGraphViz->eLayout) ) {
		case RUBY_GRAPHVIZ_NEATO:
	      neato_layout( ptRbGraphViz->g );
		  break;
		case RUBY_GRAPHVIZ_TWOPI:
		  twopi_layout( ptRbGraphViz->g );
		  break;
		case RUBY_GRAPHVIZ_DOT:
		default:
		  dot_layout( ptRbGraphViz->g );
      }
	  dotneato_write( ptRbGraphViz->g );

	  return( Qtrue );
	} else {
	  return( Qfalse );
	}
}

VALUE cGraphViz;

void Init_graphviz( void ) {
  cGraphViz = rb_define_class( "GraphViz", rb_cObject );

  rb_define_const( cGraphViz, "VERSION", rb_str_new2( RUBY_GRAPHVIZ_VERSION ) );
  
  rb_define_const( cGraphViz, "DOT",   INT2NUM( RUBY_GRAPHVIZ_DOT ) );
  rb_define_const( cGraphViz, "NEATO", INT2NUM( RUBY_GRAPHVIZ_NEATO ) );
  rb_define_const( cGraphViz, "TWOPI", INT2NUM( RUBY_GRAPHVIZ_TWOPI ) );
  
  rb_define_const( cGraphViz, "AGRAPH",          INT2NUM(AGRAPH) );
  rb_define_const( cGraphViz, "AGRAPHSTRICT",    INT2NUM(AGRAPHSTRICT) );
  rb_define_const( cGraphViz, "AGDIGRAPH",       INT2NUM(AGDIGRAPH) );
  rb_define_const( cGraphViz, "AGDIGRAPHSTRICT", INT2NUM(AGDIGRAPHSTRICT) );

  rb_define_singleton_method( cGraphViz, "new", ruby_graphviz_new2, 2 );

  rb_define_method( cGraphViz, "add_node",      ruby_graphviz_add_node,      2 );
  rb_define_method( cGraphViz, "add_edge",      ruby_graphviz_add_edge,      3 );
  rb_define_method( cGraphViz, "add_node_attr", ruby_graphviz_add_node_attr, 2 );
  rb_define_method( cGraphViz, "set_node_attr", ruby_graphviz_set_node_attr, 3 );
  rb_define_method( cGraphViz, "add_edge_attr", ruby_graphviz_add_edge_attr, 2 );
  rb_define_method( cGraphViz, "set_edge_attr", ruby_graphviz_set_edge_attr, 3 );
  rb_define_method( cGraphViz, "output",        ruby_graphviz_output,        0 );
}
